
# Lexicon, Hayes, and Cumulation
# Modules/Classes for doing Hayes type text analysis.
# Copyright (C) 2001 Marty White
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Marty White
# http://www.soc.cornell.edu/computing/
# mailto:lmw22@cornell.edu
# 380 Uris Hall, Cornell University

#==============================================================================
package Lexicon;

# A Lexicon object contains word-frequency information for a given language.

use strict;

sub new
{
  my ($pkg, $filename) = @_;
  my $self = {
    # Attributes:
    max_sentence_len => 150,
    common_by_freq => [],   # words & their u-values for the 10000 most common
    common_by_word => {},
  };
  bless $self, $pkg;
  $self->load($filename) if $filename;
  return $self;
}

sub load
{
  my ($self, $filename) = @_;
  $filename = 'common.english' if !$filename;
  local *COMMON;
  open COMMON, "< $filename"
    or die "$0: Can't open file '$filename' for reading: $!\n";
  while (my $line = <COMMON>) {
    chomp $line;
    next if $line =~ /^\s*#/;
    my @wordinfo = split(/,/, $line);
    push @{$self->{common_by_freq}}, \@wordinfo;
    $self->{common_by_word}{$wordinfo[0]} = 0;
  }
  close COMMON;
} 

#==============================================================================
package Hayes;

# A Hayes object analyzes text using Hayes' algorithm.

use strict;

sub new
{
  my ($pkg, $lexicon) = @_;
  my $self = {
    # Attributes:
    lexicon => $lexicon,
    tokencnt => 0,      # includes legit/rare and illegit, but not punct
    legittokencnt => 0, # includes rare
    raretokencnt => 0,
    raretypecnt => 0,
    illegittokencnt => 0,
    illegittypecnt => 0,
    common => {},	# count of each common word found
    rare => {},		# count of each rare word found
    illegit => {},	# count of each illegit word found
    ignored => {},
    sentencelengths => [], # histogram of sentence lengths
    currentsentencelen => 0,
    longestsentencelen => 0,
    sentencecnt => 0,
    statements => 0,
    interrogatives => 0,
    exclamation => 0,
    interruptions => 0,
    #undef max_tokencnt
    trailing_token_context => [],
    #undef idcode      # current idcode
    #undef target_idcode
    rare_std_uvalue => 1.5,
    woodpecker_limit => 6,
    woodpecker => 0,   # "Woody Woodpecker" exception tokens ignored
    std_y_intercept_1 => 3.57,
    std_y_intercept_10k => 90.13,
    STD_AREA_S_76_10000 => 85.33,  # Mignon's "scaled" std open-class area
    changed => 0,
  };
  $self->{std_slope} = Cumulation::log10_slope($self->{std_y_intercept_1},
                                               $self->{std_y_intercept_10k},
                                               1, 10000);
  return bless $self, $pkg;
}

sub reset
{
  my ($self) = @_;
  #$self->{tokencnt = 0;
  #$self->{legittokencnt = 0;
  #$self->{raretokencnt = 0;
  #$self->{raretypecnt = 0;
  #$self->{illegittokencnt = 0;
  #$self->{illegittypecnt = 0;
  #$self->{common = {};
  #$self->{rare = {};
  #$self->{illegit = {};
  #$self->{ignored = {};
  #$self->{sentencelengths = [];
  #$self->{currentsentencelen = 0;
  #$self->{longestsentencelen = 0;
  #$self->{sentencecnt = 0;
  #undef $self->{max_tokencnt};
  ##$self->{trailing_token_context = [];
  #undef $self->{idcode;
  #undef $self->{target_idcode;
  #$self->{rare_std_uvalue = 1.5;
  #$self->{woodpecker = 0;
}

sub process_file
{
  # should take a filename or file handle
  my ($self, $file) = @_;
  # call $self->process(data)
  if (ref($file)) {
    # a file handle
    #print "file handle provided\n";
    while (my $line = <$file>) {
      $self->process($line);
    }
  } else {
    # a file name
    #print "file name provided\n";
    local *INPUT;
    open INPUT, $file or die "$0: Can't open file '$file' for reading: $!\n";
      while (my $line = <INPUT>) {
        $self->process($line);
      }
    close INPUT;
  }
  # End of file implies end of sentence!
  process(".") if $self->{currentsentencelen} > 0;
  return;
}

sub process
{
  # Takes a string or array.
  my $self = shift;
  my $arg = shift;
  if (ref $arg) {
    # a reference (to an array, hopefully)
    foreach my $subarg (@$arg) {
      $self->process($subarg);
    }
  } else {
    # A scalar.

    if (!defined $arg) {
      #warn "Hayes::process(scalar) was called with an undefined value!\n";
      return;
    }

    # process id codes
    if ($arg =~ /^\s*(&&\d+)/) {
      $self->{idcode} = $1;
    }
    if ($arg =~ /^\d\d\d\d\d (\d\d\d)/) {
      $self->{idcode} = "&&$1";
      $arg = substr($arg, 9);
    }
    if (defined $self->{idcode}) {
      return if $self->{idcode} eq "&&000";   # ignore comments
      return if defined $self->{target_idcode}
             && $self->{idcode} ne $self->{target_idcode};
    }

    # Tokenize and process further.
    while ($arg =~ /((=|&&)?[0-9A-Za-z'-]+|[.?!@])/g) {
      $self->process_token($1);
    }
  }
  return;
}

sub is_too_many_tokens
{
  my ($self, $token) = @_;

  if (defined $self->{max_tokencnt}
      && $self->{legittokencnt} >= $self->{max_tokencnt})
  {
    push @{$self->{trailing_token_context}}, $token
      if @{$self->{trailing_token_context}} < 7;
    return 1;
  }
  return 0;
}

sub process_token
{
  # Add given token to on-going statistical analysis.
  # Input: $token is a plex-preprocessed token.

  my ($self, $token) = @_;

  #print "token('$token')\n";

  return if $token =~ /^&&/;
  return if $token =~ /^=?['-]+$/;

  if ($token =~ /^[.?!@]/ ) {       # end-of-sentence token?
    # End-of-sentence found, so do sentence-length analysis.
    if ($self->{currentsentencelen} > 0) {
      $self->{changed}++;
      $self->{sentencelengths}[$self->{currentsentencelen}]++;
      $self->{longestsentencelen} = $self->{currentsentencelen}
        if $self->{currentsentencelen} > $self->{longestsentencelen};
      $self->{statements}++     if $token =~ /^[.]/;
      $self->{interrogatives}++ if $token =~ /^[?]/;
      $self->{exclamation}++    if $token =~ /^[!]/;
      $self->{interruptions}++  if $token =~ /^[@]/;
      $self->{sentencecnt}++;
      $self->{currentsentencelen} = 0;
      print STDERR "." if $self->{sentencecnt} % 100 == 0; # indicate progress
    }
    $self->is_too_many_tokens($token);
    return;
  }

  return if $self->is_too_many_tokens($token);

  $self->{changed}++;

  # Not end-of-sentence, therefore sentence is even longer.
  $self->{currentsentencelen}++
    if $self->{currentsentencelen} < $self->{lexicon}{max_sentence_len};

  $self->{tokencnt}++;

  if ($token =~ /^=/) {        # Is token marked as illegit?
    $self->{illegit}{$token}++;
    $self->{illegittokencnt}++;
    return;
  }

  $self->{legittokencnt}++;
  my $lctoken = lc $token;
  if (exists $self->{lexicon}{common_by_word}{$lctoken}) {
    $self->{common}{$lctoken}++;
    $self->{commontokencnt}++;
  } else {
    if (!exists $self->{rare}{$lctoken}
      || $self->{rare}{$lctoken} < $self->{woodpecker_limit}
    ) {
      $self->{rare}{$lctoken}++;
      $self->{raretokencnt}++;
    } else {
      $self->{woodpecker}++;
      $self->{ignored}{$lctoken}++;
    }
  }

  return;
}

sub common_info
{
  # Generate an array containing info about each common word.
  # Keep this info cached for quick reuse.
  my ($self) = @_;
  if (!exists $self->{common_word_data}
    || $self->{changed_info} != $self->{changed} )
  {
    my @common_word_data;  
    my $accumulated_percentage = 0;
    foreach my $wordinfo (@{$self->{lexicon}{common_by_freq}}) {
      my ($word, $uvalue) = @$wordinfo;
      my $count = ($self->{common}{$word} || 0);

      my $proportion = eval { $count / $self->{legittokencnt} };
      if ($@)
        { $proportion = 0.0; }

      my $percentage = $proportion * 100;
      $accumulated_percentage += $percentage;
      my $ratio_to_std;
      if ($self->{legittokencnt} > 0) {
        $ratio_to_std = (1E6 * $count / $self->{legittokencnt}) / $uvalue;
      } else {
        $ratio_to_std = -1;
      }
      push @common_word_data, {
        word => $word,
        std_uvalue => $uvalue,
        count => $count,
        proportion => $proportion,
        percentage => $percentage,
        total_uvalue => $count * $uvalue,
        ratio_to_std => $ratio_to_std,
        cumulated_percentage => $accumulated_percentage,
      };
    }
    $self->{common_word_data} = \@common_word_data;
    $self->{changed_info} = $self->{changed};
    $self->{raretypecnt} = scalar(keys %{$self->{rare}});
  }
  # return a ref to the array of refs to hashes
  return $self->{common_word_data};
}

sub save_word_data
{
  my ($self, $filename, $comment) = @_;
  local *DATAFILE;

  my $data = $self->common_info();

  open DATAFILE, "> $filename"
    or die "$0: Can't open file '$filename' for writing: $!\n";
  print DATAFILE "# Word analysis raw data file\n";
  print DATAFILE "# $comment\n";
  print DATAFILE "#COUNT PROPORTION\n";
  printf DATAFILE "%d \"total legitimate tokens\"\n", $self->{legittokencnt};
  foreach my $word_datum (@$data) {
    printf DATAFILE "%d %f\n",
      $word_datum->{count}, $word_datum->{proportion};
  }
  close DATAFILE;
}

sub printable_summary
{
  my ($self, $start_rank, $stop_rank) = @_;

  my $trailing_tokens = "";
  if (defined $self->{max_tokencnt}) {
    $trailing_tokens = "; trailing tokens were:\n" . (" " x 10)
                       . join(" ", @{$self->{trailing_token_context}}) ;
  }

  if ($self->{sentencecnt} == 0) { $self->{sentencecnt} = -1; }

  return sprintf( <<"--",
General summary:
%7d tokens of all kinds (excluding punctuation)$trailing_tokens
%7d sentences
%7d   sentences ending in periods (%.1f percent)
%7d   sentences ending in question marks (%.1f percent)
%7d   sentences ending in exclamation marks (%.1f percent)
%7d   sentences ending in "@" (%.1f percent)
%7.1f tokens/sentence average
%7d tokens in longest sentence
%7d legitimate tokens
%7d rare legitimate tokens
%7d rare legitimate token types
%7d rare tokens ignored due to "Woody Woodpecker" exception
%7d illegitimate tokens
%7d illegitimate token types
--
    $self->{tokencnt},
    $self->{sentencecnt},
    $self->{statements}, 100*$self->{statements}/$self->{sentencecnt},
    $self->{interrogatives}, 100*$self->{interrogatives}/$self->{sentencecnt},
    $self->{exclamation}, 100*$self->{exclamation}/$self->{sentencecnt},
    $self->{interruptions}, 100*$self->{interruptions}/$self->{sentencecnt},
    $self->{tokencnt} / $self->{sentencecnt},
    $self->{longestsentencelen},
    $self->{legittokencnt},
    $self->{raretokencnt},
    $self->{raretypecnt},
    $self->{woodpecker},
    $self->{illegittokencnt},
    $self->{illegittypecnt},
  );
}

sub printable_micro_summary
{
  # print: word count, most over-used words, area-under-curve
  die "unimplemented"
}

sub compute_profile
{
  # Return an array containing information about each word, stripped of unused
  #   words and sorted by ratio_to_std.
  my $self = shift;
  my $data = $self->common_info();
  if (!exists $self->{profile}
    || $self->{changed_profile} != $self->{changed} )
  {
    $self->{profile} = [ sort { $a->{ratio_to_std} <=> $b->{ratio_to_std} }
                         grep { $_->{count} } @$data ];
    $self->{changed_profile} = $self->{changed};
  }
  return $self->{profile};
}

sub printable_profile
{
  my $self = shift;
  my @pp = map { sprintf "%7.2f %s\n", $_->{ratio_to_std}, $_->{word} }
               @{ $self->compute_profile() } ;
  unshift @pp, "# Ratio of actual usage to standard usage for common words:\n";
  if (wantarray) { return @pp; } else { return join("", @pp); }
}

sub printable_profile_summary
{
  my ($self) = @_;
  my @pp;

  eval {
  @pp = "Most overused words are: " .
           join( ", ",
             map { sprintf "%s(%.0f)", $_->{word}, $_->{ratio_to_std} }
               reverse
                 splice @{ $self->compute_profile() }, -8
           ) . "\n";
  };
  if ($@) { @pp = ""; }

  if (wantarray) { return @pp; } else { return join("", @pp); }
}

sub save_sentence_histogram
{
  my ($self, $filename, $comment) = @_;
  local *SENTENCES;

  open SENTENCES, "> $filename"
    or die "$0: Can't open file '$filename' for writing: $!\n";
  print SENTENCES "# Sentence lengths histogram data file\n";
  print SENTENCES "# $comment\n";
  print SENTENCES "#LEN COUNT\n";
  for (my $i=1; $i<=$self->{longestsentencelen}; $i++) {
    printf SENTENCES " %3d %5d\n", $i, ($self->{sentencelengths}[$i] || 0);
  }
  close SENTENCES;
}

sub printable_rare
{
  my ($self) = @_;
  my @pp;
  $self->{raretypecnt} = scalar(keys %{$self->{rare}});
  push @pp, "# $self->{raretokencnt} rare legitimate tokens, $self->{raretypecnt} rare legitimate token types:\n";
  push @pp, map { sprintf "%5d  $_\n", $self->{rare}{$_} }
            sort keys %{$self->{rare}};
  if (wantarray) { return @pp; } else { return join("", @pp); }
}

sub printable_ignored
{
  my ($self) = @_;
  my (@pp, $n);
  $n = keys %{$self->{ignored}};
  push @pp,"# $self->{woodpecker} rare tokens ($n token types) were ignored:\n";
  push @pp, map { sprintf "%5d  $_\n", $self->{ignored}{$_} }
            sort keys %{$self->{ignored}};
  if (wantarray) { return @pp; } else { return join("", @pp); }
}

sub printable_illegit
{
  my ($self) = @_;
  my @pp;
  $self->{illegittypecnt} = scalar(keys %{$self->{illegit}});
  push @pp, "# $self->{illegittokencnt} illegitimate tokens, $self->{illegittypecnt} illegitimate token types:\n";
  push @pp, map { sprintf "%5d  $_\n", $self->{illegit}{$_} }
            sort keys %{$self->{illegit}};
  if (wantarray) { return @pp; } else { return join("", @pp); }
}

1;

#==============================================================================
package Cumulation;

# A Cumulation object computes cumluated percentages and areas-under-the-curve.

sub new
{
  my ($pkg, $hayes, $start_rank, $stop_rank) = @_;

  $start_rank = 1 if !$start_rank;
  $stop_rank = 10000 if !$stop_rank;  # or $stop_rank = @$data

  my $cumulation = {
    hayes => $hayes,
    start_rank => $start_rank,
    stop_rank => $stop_rank,
    include_rare => ($stop_rank >= 10000),
    word_type_count => 0,
    word_count => 0,
    sum_uvalues => 0,
    sum_percentage => 0,
    x1 => 0,
    y1 => 0,
  };
  bless $cumulation, $pkg;

  $cumulation->compute();

  return $cumulation;
}

sub log10
{
  return log($_[0]) / log(10);                                                  }

sub log10_slope
{ # Compute slope of a line, after converting x ordinates to log10 scale.
  # typical use: slope(1.4, 81.8, 76, 10000)
  my ($y1, $y2, $x1, $x2) = @_;
  return ($y2 - $y1) / ( log10($x2) - log10($x1) );
}

sub log10_trapezoid_area
{
  # Compute area of trapezoid defined by
  #   logarithmic x axis and the provided line.
  my ($y1, $y2, $x1, $x2) = @_;

  ($y1,$y2) = ($y2,$y1) if $y1 > $y2;  # Ensure y1 <= y2

  # area = box base + triangle top
  my $width = log10($x2) - log10($x1);
  return $width * ($y2 - $y1) / 2   # triangle = width * height / 2
       + $width * $y1;              # rectangle = width * height of lower side
}

sub std_intercept
{ # Compute the y intercept of the standard model at the given x ordinate.
  my ($self, $x) = @_;

  # y = mx + b
  my $y = $self->{hayes}{std_slope} * log10($x)
        + $self->{hayes}{std_y_intercept_1};
  return $y;
}

sub reset
{
  my ($self) = @_;
  $self->{word_type_count} = 0;
  $self->{word_count} = 0;
  $self->{sum_uvalues} = 0;
  $self->{sum_percentage} = 0;
  $self->{x1} = 0;
  delete $self->{triangles_area};
  delete $self->{trapezoids_area};
  delete $self->{left_boxes_area};
  $self->{mean_uvalue} = 0;
}

sub compute
{
  # Examine the list of common words for this sample and compute area under
  # the curve & other related statistics, starting and stopping at given ranks.

  my ($self, $callback) = @_;

  # $common_data is a ref to an array of refs to hashes
  my $common_data = $self->{hayes}->common_info();

  $self->reset();

  #$self->{sum_percentage} = $$common_data[$self->{start_rank}-1]->{accumulated_percentage};

  for (my $x=$self->{start_rank}; $x <= $self->{stop_rank}; $x++) {
    my $datum = $$common_data[$x-1];  # $datum is a ref to a hash

    if (!defined $self->{left_boxes_area}) {#Can't add trapezoid until 2nd iter
      # First pass setup.

      # Initialize area accumulators.
      $self->{triangles_area}  = 0.0;
      $self->{trapezoids_area} = 0.0;
      $self->{left_boxes_area} = 0.0;

      $self->{x1} = log10($x);    # Record coords of 1st point.
    } else {
      # Nth pass, where N>1.
      my $x2 = log10($x);  # Compute right side x-ordinate.
      my $width = $x2 - $self->{x1};
      # triangle = width * height / 2
      # triangle includes only the bit at the top
      my $triangle = $width * $datum->{percentage} / 2;
      # rectangle = width * height of lower side
      # rectangle for this rank goes down to y = start cumulation  !!
      my $left_box = $width * $self->{sum_percentage};
      # rectangle = width * height of upper side
      #my $right_box = $width * ($self->{sum_percentage}+$datum->{percentage});

      $self->{triangles_area}  += $triangle;             # possibly pointless.
      $self->{left_boxes_area} += $left_box;             # seemingly what we want.

      # Record coords of this point for next pass.
      $self->{x1} = $x2;
    }
    $self->{trapezoids_area} = $self->{triangles_area} + $self->{left_boxes_area};
    $self->{sum_percentage} += $datum->{percentage};
    $self->{word_type_count}++ if $datum->{count};
    $self->{word_count} += $datum->{count};
    $self->{sum_uvalues} += $datum->{total_uvalue};
    #if ($x == 75 && $self->{start_rank} == 1) {
    #  $self->{hayes}{sum_percentage_75} = $self->{sum_percentage};
    #}

    # Consider instead storing an array of sum_percentages
    &$callback($self, $x, $datum) if $callback;
  }

  if ($self->{include_rare}) {
    $self->{sum_uvalues} +=
        $self->{hayes}{raretokencnt} * $self->{hayes}{rare_std_uvalue};
    $self->{word_count} += $self->{hayes}{raretokencnt};
    $self->{word_type_count} += $self->{hayes}{raretypecnt};
    # Note that area is not adjusted here.
  }

  if ($self->{word_count} == 0) {
    $self->{mean_uvalue} = -1; # really, undef
  } else {
    $self->{mean_uvalue} =
        $self->{sum_uvalues} / $self->{word_count};
  }

  $self->{starting_cumulation} = 0;
  $self->{starting_cumulation} =
      $$common_data[ $self->{start_rank} - 2 ]{cumulated_percentage}
    if $self->{start_rank} > 1;
  $self->{area_under_start_rank} =  # width * height:
    ( log10($self->{stop_rank}) - log10($self->{start_rank}) ) *
    $self->{starting_cumulation};

  $self->{tall_boxes_area} = $self->{left_boxes_area} + $self->{area_under_start_rank};

  if ($self->{start_rank} == 76 && $self->{stop_rank} == 10000 && $self->{starting_cumulation} != 100) {
    die if $self->{starting_cumulation} == 100;
    $self->{mignon_lex} =
      #85.3 - ( $self->{area} * 100 / (100 - $self->{hayes}{sum_percentage_75}))
      $self->{hayes}{STD_AREA_S_76_10000}
        - ( $self->{left_boxes_area} * 100 /
            (100 - $self->{starting_cumulation}) );
  }

  return $self;
}

sub save
{
  my ($self, $filename, $comment) = @_;
  $comment = "" if !$comment;

  local *CUMULATION;
  open CUMULATION, ">$filename"
    or die "$0: Can't open file '$filename' for writing: $!";
  print CUMULATION "# Cumulative Proportion Distribution data file\n";
  print CUMULATION "# $comment\n";
  print CUMULATION "# For the words from rank $self->{start_rank} to $self->{stop_rank}:\n";
  $self->compute( sub {
    my ($c, $x, $d) = @_;
    printf CUMULATION " %5d %5.2f\n", $x, $c->{sum_percentage};
  });
  close CUMULATION;
}

sub printable_report
{
  my ($self) = @_;
  my ($format_header, $last_reported_rank, @records);
  my $format_record = "%5d%s %6d %5d  %7.4f%%  %6.2f%%  %s\n";

  ($format_header = $format_record) =~ s/([.][0-9])?[dfg]/s/g;
  $format_header =~ s/%([0-9])s%%/ "%" . ($1+1) . "s" /ge;
  push @records, sprintf("#Word listing for cumulative proportion distribution from $self->{start_rank} to $self->{stop_rank}:\n");
  push @records, sprintf($format_header,
    "#RANK", " ", "UVALUE", "N", "PERCENT", "CUM.%", "WORD");

  $last_reported_rank = 0;
  $self->compute( sub {
    my ($c, $x, $d) = @_;
    if ($d->{count} > 0 || grep($_==$x,
      (1,2,3,4,5,6,7,8,9,10,15,20,25,35,50,60,75,100,150,200,250,300,400,500,
       600,700,800,900,1000,1200,1400,1600,1800,2000,2500,3000,3500,4000,
       4500,5000,5500,6000,6500,7000,7500,8000,8500,9000,9500,10000))
    ) {
      my $flag = (($x-1 == $last_reported_rank) ? ' ' : ':');
      $last_reported_rank = $x;
      push @records, sprintf $format_record,
        $x, $flag, $d->{std_uvalue}, $d->{count}, $d->{percentage},
        $c->{sum_percentage}, $d->{word};
    }
  });
  if (wantarray) {
    return @records;
  } else {
    return join("",@records);
  }
}

sub printable_summary
{
  my ($self) = @_;

  # Note that areas *never* include rare words.

  my $y_intercept_start = $self->std_intercept($self->{start_rank});
  my $y_intercept_stop  = $self->std_intercept($self->{ stop_rank});
  my $std_width_range = log10($self->{stop_rank}) - log10($self->{start_rank});
  my $std_area_triangle_range = ($y_intercept_stop - $y_intercept_start)
                               * $std_width_range / 2;
  my $std_area_rectangle_range = $y_intercept_start * $std_width_range;
  my $std_area_trapezoid_range = $std_area_rectangle_range + $std_area_triangle_range;

  my $summary = "";

  $summary .= sprintf( <<"--",
For the words from rank $self->{start_rank} to $self->{stop_rank}%s:
  token count = $self->{word_count}
  token types = $self->{word_type_count}
  mean U-value = %.4f
--
    ($self->{include_rare}?" plus rare legitimate words":""),
    $self->{mean_uvalue},
  );

 if (!$self->{include_rare}) {

  $summary .= sprintf("  Area under STANDARD model [line from (%d,%.2f%%) to (%d,%.2f%%)]:\n",
    $self->{start_rank},
    $y_intercept_start,
    $self->{stop_rank},
    $y_intercept_stop,
  );

  $summary .= sprintf("    area of triangle                      :   %8.4f\n", $std_area_triangle_range);
  $summary .= sprintf("    area of rectangle [y=0.0%% to y=%.2f%%]: + %8.4f\n",
                             $y_intercept_start, $std_area_rectangle_range);
  $summary .= sprintf("    area of trapezoid                     : = %8.4f\n", $std_area_trapezoid_range);
  $summary .= sprintf("  Areas for THIS TEXT:\n");
  $summary .= sprintf("    starting cumulation = %.2f%%\n", $self->{starting_cumulation} );
  $summary .= sprintf("    total area of little triangles               :   %8.4f\n",
                             $self->{triangles_area});
  $summary .= sprintf("    total area of boxes above starting cumulation: + %8.4f\n",
                             $self->{left_boxes_area});
  $summary .= sprintf("    total area of trapezoids above               : = %8.4f\n",
                             $self->{trapezoids_area});
  $summary .= sprintf("    area of big box under starting cumulation: %.4f\n",
                             $self->{area_under_start_rank});
  $summary .= sprintf("    area of boxes down to y axis: %.4f\n", $self->{tall_boxes_area});
  $summary .= sprintf("  Standard area triangle minus area of upper little boxes: %.4f\n",
                             $std_area_triangle_range - $self->{left_boxes_area} );
  $summary .= sprintf("  Standard area rectangle minus area of big box under start cumulation: %.4f\n",
                             $std_area_rectangle_range - $self->{area_under_start_rank} );
  $summary .= sprintf("  Standard area trapezoid minus area of boxes down to y axis: %.4f\n",
                             $std_area_trapezoid_range - $self->{tall_boxes_area} );


  if (defined $self->{mignon_lex}) {
    $summary .= sprintf
      "  Mignon's LEX score = %.2f - [ %.4f * 100 / (100 - %.4f) ] = %.2f\n",
      $self->{hayes}{STD_AREA_S_76_10000},
      $self->{left_boxes_area},
      $self->{starting_cumulation},
      $self->{mignon_lex}
      ;
  }

 }

  return $summary;
}

1;

#==============================================================================
package Plexer;

# Provide preprocessing & Hayes markup functions.

sub new
{ 
  my ($pkg) = @_;
  my $self = { 
    # Attributes:
    idcode => "",
    antispam => 1,
    unhtml => 1,
  };
  bless $self, $pkg;
  return $self;
}

# [ UNIMPLEMENTED ]

1;

